(*  Title:      HOL/Bali/DecimalExample.thy
    ID:         $Id: Example.thy,v 1.4 2002/02/22 10:26:46 schirmer Exp $
    Author:     Leonor Prensa
    License:    GPL (GNU GENERAL PUBLIC LICENSE)
*)

header {* Another example Bali program: Decimal representation in Java
for Smart Cards *}

theory DecimalExample = Eval + WellForm:

text {*
The following example Bali program includes:
\begin{itemize}
\item class and interface declarations with inheritance, hiding of fields,
    overriding of methods (with refined result type), array type,
\item method call (with dynamic binding), parameter access, return expressions,
\item expression statements, sequential composition, literal values, 
    local assignment, local access, field assignment, type cast,
\item exception generation and propagation, try and catch statement, throw 
      statement
\item instance creation and (default) static initialization
\end{itemize}
\begin{verbatim}
package java_lang

public class Decimal extends Object {
  public static final short PRECISION;
  private DecimalException decimal_exception;
  private short intPart;
  private short decPart;
  public Decimal addOpt (Decimal) throws DecimalException;
  private void addOpt (short, short);
  public Decimal setValue (Decimal);
  public Decimal oppose ();
}

public Decimal oppose() {
  intPart = (short) -intPart;
  decPart = (short) -decPart;
  return this;
}

private DecimalException decimal_exception
   = new DecimalException ();

public Decimal setValue (Decimal d)
  throws DecimalException {
    return setValue (d.getIntPart(), d.getDecPart());
}

private void addOpt (short e, short f) {
  intPart += e;
  decPart += f;
  if (decPart \<le> -PRECISION ) {
      decPart += PRECISION;
      intPart --;
    }
  else if (PRECISION \<le> decPart ) {
           decPart -= PRECISION;
           intPart ++; 
  }
}

\end{verbatim}
*}

thm widen.null

declare widen.null [intro]

lemma wf_fdecl_def2: "\<And>fd. wf_fdecl G P fd = is_acc_type G P (type (snd fd))"
apply (unfold wf_fdecl_def)
apply (simp (no_asm))
done

declare wf_fdecl_def2 [iff]


section "type and expression names"

(** unfortunately cannot simply instantiate tnam **)
datatype tnam_  = Decimal_ | DecimalException_ | Main_
datatype vnam_  = intPart_ | decPart_ | decimal_exception_ | PRECISION_ 
                  | vnd_ | vne_ | vnf_
datatype label_ = lab1_

consts
  tnam_ :: "tnam_  \<Rightarrow> tnam"
  vnam_ :: "vnam_  \<Rightarrow> vname"
  label_:: "label_ \<Rightarrow> label"

axioms  (** tnam_, vnam_ and label are intended to be isomorphic 
            to tnam, vname and label **)

  inj_tnam_  [simp]: "(tnam_  x = tnam_  y) = (x = y)"
  inj_vnam_  [simp]: "(vnam_  x = vnam_  y) = (x = y)"
  inj_label_ [simp]: "(label_ x = label_ y) = (x = y)"
  
  
  surj_tnam_:  "\<exists>m. n = tnam_  m"
  surj_vnam_:  "\<exists>m. n = vnam_  m"
  surj_label_:" \<exists>m. n = label_ m"

syntax
  Decimal :: qtname
  DecimalException :: qtname
  Main   :: qtname
  intPart :: vname
  decPart :: vname
  decimal_exception :: vname
  PRECISION   :: vname
  vnd :: vname
  vne :: vname
  vnf :: vname
  lab1:: label


translations
  "Decimal" == "\<lparr>pid=java_lang,tid=TName (tnam_ Decimal_)\<rparr>"
  "DecimalException" == "\<lparr>pid=java_lang,tid=TName (tnam_ DecimalException_)\<rparr>"
  "Main" == "\<lparr>pid=java_lang,tid=TName (tnam_ Main_)\<rparr>"
  "intPart" == "(vnam_ intPart_)"
  "decPart" == "(vnam_ decPart_)"
  "decimal_exception" == "(vnam_ decimal_exception_)"
  "PRECISION" == "(vnam_ PRECISION_)"
  "vnd" == "(vnam_ vnd_)"
  "vne" == "(vnam_ vne_)"
  "vnf" == "(vnam_ vnf_)"
  "lab1" == "label_ lab1_"

constdefs
  intPart_fspec :: fspec
  "intPart_fspec \<equiv> (intPart, Decimal)"

  decPart_fspec :: fspec
  "decPart_fspec \<equiv> (decPart, Decimal)"

  decimal_exception_fspec :: fspec
  "decimal_exception_fspec \<equiv> (decimal_exception, DecimalException)"

lemma neq_Decimal_Object [simp]: "Decimal\<noteq>Object"
by (simp add: Object_def)

lemma neq_DecimalException_Object [simp]: "DecimalException\<noteq>Object"
by (simp add: Object_def)

lemma neq_Main_Object [simp]: "Main\<noteq>Object"
by (simp add: Object_def)

lemma neq_Decimal_SXcpt [simp]: "Decimal\<noteq>SXcpt xn"
by (simp add: SXcpt_def)

lemma neq_DecimalException_SXcpt [simp]: "DecimalException\<noteq>SXcpt xn"
by (simp add: SXcpt_def)

lemma neq_Main_SXcpt [simp]: "Main\<noteq>SXcpt xn"
by (simp add: SXcpt_def)

section "field declarations" 

constdefs
 PRECISION_fdecl:: fdecl
 "PRECISION_fdecl \<equiv> (PRECISION, \<lparr>access=Public, static=True, type=PrimT Integer\<rparr>)" 

 decimal_exception_fdecl:: fdecl
 "decimal_exception_fdecl \<equiv> (decimal_exception, \<lparr>access=Private, static=False, type=Class DecimalException\<rparr>)"

 intPart_fdecl:: fdecl
 "intPart_fdecl \<equiv> (intPart, \<lparr>access=Private, static=False, type=PrimT Integer\<rparr>)"

 decPart_fdecl :: fdecl
 "decPart_fdecl \<equiv> (decPart, \<lparr>access=Private, static=False, type=PrimT Integer\<rparr>)"


section "classes and interfaces"

defs
  Object_mdecls_def: "Object_mdecls \<equiv> []"
  SXcpt_mdecls_def:  "SXcpt_mdecls  \<equiv> []"

consts
  setValue :: mname
  oppose :: mname
  addOpt :: mname

constdefs
  setValue_sig :: sig
 "setValue_sig   \<equiv> \<lparr>name=setValue,parTs=[Class Decimal]\<rparr>"
  
  setValue_mhead :: mhead
 "setValue_mhead \<equiv> \<lparr>access=Public,static=False,pars=[vnd],resT=Class Decimal\<rparr>"

  Priv_setValue_sig :: sig
 "Priv_setValue_sig   \<equiv> \<lparr>name=setValue,parTs=[PrimT Integer, PrimT Integer]\<rparr>"
  
  Priv_setValue_mhead :: mhead
 "Priv_setValue_mhead \<equiv> \<lparr>access=Private,static=False,pars=[vnd],resT=Class Decimal\<rparr>"

  oppose_sig :: sig
 "oppose_sig   \<equiv> \<lparr>name=oppose,parTs=[]\<rparr>"
  
  oppose_mhead :: mhead
 "oppose_mhead \<equiv> \<lparr>access=Public,static=False,pars=[],resT=Class Decimal\<rparr>"

  addOpt_sig :: sig
 "addOpt_sig   \<equiv> \<lparr>name=addOpt,parTs=[Class Decimal]\<rparr>"
  
  addOpt_mhead :: mhead
 "addOpt_mhead \<equiv> \<lparr>access=Public,static=False,pars=[vnd],resT=Class Decimal\<rparr>"

  Priv_addOpt_sig :: sig
 "Priv_addOpt_sig   \<equiv> \<lparr>name=addOpt,parTs=[PrimT Integer, PrimT Integer]\<rparr>"
  
  Priv_addOpt_mhead :: mhead
 "Priv_addOpt_mhead \<equiv> \<lparr>access=Private,static=False,pars=[vne, vnf],resT=PrimT Void\<rparr>"

constdefs
  Decimal_oppose :: mdecl
 "Decimal_oppose \<equiv> (oppose_sig, mhead.extend oppose_mhead 
                    \<lparr>mbody = \<lparr>lcls=[],
                              stmt= Expr ({Decimal, Decimal, True}this..intPart:=
                         UnOp UMinus (Acc ({Decimal, Decimal, True}this..intPart)));;
                           Expr ({Decimal, Decimal, True}this..decPart:=
                         UnOp UMinus (Acc ({Decimal, Decimal, True}this..decPart)));;
                          Return this \<rparr> \<rparr>)"

(*  
constdefs

Decimal_setValue :: mdecl
 "Decimal_setValue \<equiv> (setValue_sig, mhead.extend setValue_mhead \<lparr>mbody=\<lparr>lcls=[],stmt=Return (Call )\<rparr>\<rparr>)"

  Decimal_Priv_setValue :: mdecl
 "Decimal_Priv_setValue \<equiv> (Priv_setValue_sig, \<lparr>access=Private,static=False,pars=[d],resT=Class Decimal,
                        mbody=\<lparr>lcls=[],stmt=Return (!!z)\<rparr>\<rparr>)"

  Decimal_addOpt :: mdecl
 "Decimal_addOpt \<equiv> (addOpt_sig, \<lparr>access=Public,static=False,pars=[z],resT=Class Decimal,
                        mbody=\<lparr>lcls=[],stmt=Return (!!z)\<rparr>\<rparr>)"

  Decimal_Priv_addOpt :: mdecl
 "Decimal_Priv_addOpt \<equiv> (Priv_addOpt_sig, \<lparr>access=Private,static=False,pars=[e, f],resT=Void,
                        mbody=\<lparr>lcls=[],stmt=Return (!!z)\<rparr>\<rparr>)"
*)

constdefs
  
DecimalCl :: class
"DecimalCl \<equiv> \<lparr>access=Public,
           cfields=[PRECISION_fdecl,decimal_exception_fdecl,intPart_fdecl,decPart_fdecl],
           methods=[Decimal_oppose],
           init=Skip,
           super=Object,
           superIfs=[]\<rparr>"
(*  
DecimalExceptionCl  :: class
"DecimalExceptionCl  \<equiv> \<lparr>access=Public,
           cfields=[(vee, \<lparr>access=Public,static=False,type= PrimT Integer\<rparr>)], 
           methods=[DecimalException_foo],
           init=Skip,
           super=Decimal,
           superIfs=[]\<rparr>"
*)

MainCl :: class
"MainCl \<equiv> \<lparr>access=Public,
           cfields=[], 
           methods=[], 
           init=Skip,
           super=Object,
           superIfs=[]\<rparr>"
(* The "main" method is modeled seperately (see tprg) *)

constdefs
(*  
  HasFooInt :: iface
 "HasFooInt \<equiv> \<lparr>access=Public,imethods=[(foo_sig, foo_mhead)],isuperIfs=[]\<rparr>"
*)
  Ifaces ::"idecl list"
 "Ifaces \<equiv> []"

 "Classes" ::"cdecl list"
 "Classes \<equiv> [(Decimal,DecimalCl),(Main,MainCl)]@standard_classes"

lemmas table_classes_defs = 
     Classes_def standard_classes_def ObjectC_def SXcptC_def
(*
lemma table_ifaces [simp]: "table_of Ifaces = empty(HasFoo\<mapsto>HasFooInt)"
apply (unfold Ifaces_def)
apply (simp (no_asm))
done
*)
lemma table_classes_Object [simp]: 
 "table_of Classes Object = Some \<lparr>access=Public,cfields=[]
                                 ,methods=Object_mdecls
                                 ,init=Skip,super=arbitrary,superIfs=[]\<rparr>"
apply (unfold table_classes_defs)
apply (simp (no_asm) add:Object_def)
done

lemma table_classes_SXcpt [simp]: 
  "table_of Classes (SXcpt xn) 
    = Some \<lparr>access=Public,cfields=[],methods=SXcpt_mdecls,
            init=Skip,
            super=if xn = Throwable then Object else SXcpt Throwable,
            superIfs=[]\<rparr>"
apply (unfold table_classes_defs)
apply (induct_tac xn)
apply (simp add: Object_def SXcpt_def)+
done
(*
lemma table_classes_HasFoo [simp]: "table_of Classes HasFoo = None"
apply (unfold table_classes_defs)
apply (simp (no_asm) add: Object_def SXcpt_def)
done
*)
lemma table_classes_Decimal [simp]: "table_of Classes Decimal = Some DecimalCl"
apply (unfold table_classes_defs )
apply (simp (no_asm) add: Object_def SXcpt_def)
done
(*
lemma table_classes_DecimalException [simp]: "table_of Classes DecimalException = Some DecimalExceptionCl"
apply (unfold table_classes_defs )
apply (simp (no_asm) add: Object_def SXcpt_def)
done
*)
lemma table_classes_Main [simp]: "table_of Classes Main = Some MainCl"
apply (unfold table_classes_defs )
apply (simp (no_asm) add: Object_def SXcpt_def)
done

section "program"

syntax
  tprg :: prog

translations
  "tprg" == "\<lparr>ifaces=Ifaces,classes=Classes\<rparr>"

term "{Decimal,ClassT Decimal, IntVir}x\<cdot>oppose ({[]}[])"


end
